﻿using Ptv.XServer.Controls.Map;
using Ptv.XServer.Controls.Map.Canvases;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;

namespace SelectionWinForms
{
    public class SelectInteractor : WorldCanvas
    {
        private readonly ObservableCollection<FrameworkElement> shapes;
        private readonly MapView map;

        private Point g1;
        private SelectMode selectMode = SelectMode.None;
        private List<Point> polyPoints;
        private Polygon dragPolygon;

        public ObservableCollection<FrameworkElement> SelectedElements = new ObservableCollection<FrameworkElement>();

        public SelectInteractor(MapView map, ObservableCollection<FrameworkElement> shapes) : base(map)
        {
            this.map = map;
            this.shapes = shapes;

            map.MouseLeftButtonDown += map_MouseLeftButtonDown;
            map.MouseMove += map_MouseMove;
            map.MouseLeftButtonUp += map_MouseLeftButtonUp;

            SetZIndex(this, 999999999);
            SelectedElements.CollectionChanged += SelectedElements_CollectionChanged;
        }

        private void map_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            if (dragPolygon != null)
                Children.Remove(dragPolygon);

            g1 = CanvasToPtvMercator(e.GetPosition(this));

            if ((Keyboard.Modifiers & ModifierKeys.Alt) != 0)
            {
                selectMode = SelectMode.Polygon;
                polyPoints = new List<Point> { g1 };
                dragPolygon = new Polygon
                {
                    Fill = new SolidColorBrush(Color.FromArgb(0x3e, 0xa0, 0x00, 0x00)),
                    Stroke = new SolidColorBrush(Color.FromArgb(0x55, 0xff, 0x00, 0x00))
                };

                Children.Add(dragPolygon);
                e.Handled = true;
            }
            else
                selectMode = SelectMode.Click;
        }

        private void map_MouseMove(object sender, MouseEventArgs e)
        {
            if (selectMode != SelectMode.Polygon) return;

            polyPoints.Add(CanvasToPtvMercator(e.GetPosition(this)));

            dragPolygon.Points.Clear();
            polyPoints.ForEach(point => dragPolygon.Points.Add(PtvMercatorToCanvas(point)));

            e.Handled = true;
        }

        private void map_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            if ((Keyboard.Modifiers & ModifierKeys.Shift) == 0)
                SelectedElements.Clear();

            switch (selectMode)
            {
                case SelectMode.Click:
                    {
                        var selectedShapes =
                            from shape in this.shapes
                            let g2 = e.GetPosition(shape)
                            let x = shape.InputHitTest(g2)
                            let pin = GetLogicalParent<Ptv.XServer.Controls.Map.Symbols.Pin>(x as DependencyObject)
                            where pin == shape
                            select shape;
                        foreach (var shape in selectedShapes)
                        {
                            if (SelectedElements.Contains(shape))
                                SelectedElements.Remove(shape);
                            else
                                SelectedElements.Add(shape);
                        }
                        break;
                    }
                case SelectMode.Polygon:
                    {
                        CanvasToPtvMercator(e.GetPosition(map));
                        polyPoints.Add(polyPoints[0]);

                        var selectedShapes =
                            from shape in shapes
                            let pp1 = (new Point(GetLeft(shape), GetTop(shape)))
                            let x = dragPolygon.InputHitTest(pp1)
                            where x == dragPolygon
                            where !SelectedElements.Contains(shape)
                            select shape;
                        foreach (var shape in selectedShapes)
                            SelectedElements.Add(shape);

                        Children.Remove(dragPolygon);
                        dragPolygon = null;
                    }
                    break;
            }

            selectMode = SelectMode.None;
        }

        private void SelectedElements_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (e.NewItems != null)
            {
                foreach (FrameworkElement shape in e.NewItems)
                    (shape as Ptv.XServer.Controls.Map.Symbols.Pin).Color = Colors.Red;
            }

            if (e.OldItems != null)
            {
                foreach (FrameworkElement shape in e.OldItems)
                    (shape as Ptv.XServer.Controls.Map.Symbols.Pin).Color = Colors.Blue;
            }

            if (e.Action == System.Collections.Specialized.NotifyCollectionChangedAction.Reset)
            {
                foreach(var shape in shapes)
                    (shape as Ptv.XServer.Controls.Map.Symbols.Pin).Color = Colors.Blue;
            }
        }

        public override void Update(UpdateMode updateMode)
        {
            foreach (var e in Children.OfType<Ellipse>())
                e.RenderTransform = new ScaleTransform(map.CurrentScale, map.CurrentScale);
        }

        public void Remove()
        {
            map.MouseLeftButtonUp -= map_MouseLeftButtonUp;
            map.MouseLeftButtonDown -= map_MouseLeftButtonDown;
            map.MouseMove -= map_MouseMove;

            map.GeoCanvas.Children.Remove(this);
        }

        public static T GetLogicalParent<T>(DependencyObject p_oElement) where T : DependencyObject
        {
            DependencyObject oParent = p_oElement;
            Type oTargetType = typeof(T);

            if (p_oElement == null)
            {
                return null as T;
            }

            do
            {
                oParent = LogicalTreeHelper.GetParent(oParent);
            }
            while (!(oParent == null || oParent.GetType() == oTargetType || oParent.GetType().IsSubclassOf(oTargetType)));

            return oParent as T;
        }
    }

    public enum SelectMode
    {
        None,
        Click,
        Polygon
    }
}